#!/usr/bin/env python
# ******************************************************************************
# Copyright 2019 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
A set of functions to convert a Keras (tf.keras) model to a new
equivalent model with different characteristics. Then, the new model
can be quantized.

"""
from tensorflow.keras.layers import (Conv2D, SeparableConv2D, Dense, MaxPool2D,
                                     GlobalAveragePooling2D, BatchNormalization)
from tensorflow.keras.models import (load_model, model_from_json, clone_model)
from .quantization_layers import (QuantizedConv2D, QuantizedSeparableConv2D,
                                  QuantizedDense)
from .transforms.sequential import compute_BN_folded_weights
from .cnn2snn_objects import cnn2snn_objects


def invert_batchnorm_pooling(model):
    """Returns a new model where pooling and batchnorm layers are inverted.
    From a Keras model where pooling layers precede batch normalization
    layers, this function places the BN layers before pooling layers. This
    is the first step before folding BN layers into neural layers.
    Note: inversion of layers is equivalent only if the gammas of BN layers
    are positive. The function raises an error if not.

    Args:
        model (:obj:`tf.keras.Model`): a tf.keras model.
    Returns:
        :obj:`tf.keras.Model`: a keras.Model.
    """

    # Maps between successive pooling->batchnorm layers. These pairs will be
    # inverted when cloning the model.
    pool2bn_map = {}
    bn2pool_map = {}

    for layer in model.layers:
        # We map BatchNormalization layers that have only one inbound layer
        # being a MaxPool2D or GlobalAveragePooling2D.
        if (isinstance(layer, BatchNormalization) and
                len(layer.inbound_nodes) == 1 and
                isinstance(layer.inbound_nodes[0].inbound_layers,
                           (MaxPool2D, GlobalAveragePooling2D))):
            bn2pool_map[layer] = layer.inbound_nodes[0].inbound_layers
            pool2bn_map[layer.inbound_nodes[0].inbound_layers] = layer

    def replace_layer(layer):
        if layer in pool2bn_map:
            # Replace pooling layer with the corresponding BN layer
            layer_bn = pool2bn_map[layer]
            config_bn = layer_bn.get_config()
            if isinstance(layer, GlobalAveragePooling2D):
                config_bn['axis'] = [-1]
            return layer_bn.__class__.from_config(config_bn)
        if layer in bn2pool_map:
            # Replace BN layer with the corresponding pooling layer
            layer_pool = bn2pool_map[layer]
            return layer_pool.__class__.from_config(layer_pool.get_config())
        return layer.__class__.from_config(layer.get_config())

    new_model = clone_model(model, clone_function=replace_layer)

    for layer in model.layers:
        new_model.get_layer(layer.name).set_weights(layer.get_weights())

    return new_model


def fold_batch_norms(model):
    """Returns a new model where batchnorm layers are folded into
    previous neural layers.

    From a Keras model where BN layers follow neural layers, this
    function removes the BN layers and updates weights and bias
    accordingly of the preceding neural layers. The new model is
    strictly equivalent to the previous one.

    Args:
        model (:obj:`tf.keras.Model`): a Keras model.
    Returns:
        :obj:`tf.keras.Model`: a tf.keras.Model.
    """

    # Copy model in order to modify before cloning.
    model_copy = clone_model(model)
    model_copy.set_weights(model.get_weights())

    # Get BN layers to fold, mapping with the preceding neural layer and the
    # following layer if present.
    map_neural_layer_to_bn, map_bn_to_next_layer = _find_batchnorms_to_fold(
        model_copy)

    # Update model before cloning: remove BN layers, update inbound nodes and
    # output layers
    _prepare_model_to_fold_BN(model_copy, map_neural_layer_to_bn,
                              map_bn_to_next_layer)

    # Clone model (BN layers will disappear)
    def replace_layer(layer):
        config = layer.get_config()
        # Set use_bias=True in neural layers where BN layers are folded, to
        # accept new computed weights
        if layer in map_neural_layer_to_bn:
            config['use_bias'] = True
        return layer.__class__.from_config(config)

    model_foldBN = clone_model(model_copy, clone_function=replace_layer)

    # Set weights in the new model
    for layer in model_copy.layers:
        if layer not in map_neural_layer_to_bn:
            model_foldBN.get_layer(layer.name).set_weights(layer.get_weights())
        else:
            # Set new weights
            layer_bn = map_neural_layer_to_bn[layer]
            new_weights = compute_BN_folded_weights(layer, layer_bn)[0]
            model_foldBN.get_layer(layer.name).set_weights(new_weights)

    return model_foldBN


def _find_batchnorms_to_fold(model):
    """Detect BatchNormalization layers that can be folded.

    We limit to BN layers that follow supported neural layer types. Moreover,
    we only fold BN layers with one inbound/outbound node, in order to avoid
    layers with multiple nodes (e.g. layers used multiple times in the model
    or in other models).

    The function returns two maps (dict):
    - one between the neural layer and its BN layer to be folded
    - one between the BN layer and its following layer (it if exists)

    """

    supported_layers = (Conv2D, SeparableConv2D, Dense, QuantizedConv2D,
                        QuantizedSeparableConv2D, QuantizedDense)

    # Map between a neural layer and its following BN layer, and map between
    # the BN layer and its next layer.
    map_neural_layer_to_bn = {}
    map_bn_to_next_layer = {}

    # Find triplet "neural layer -> BN -> next layer"
    for layer in model.layers:
        # Find a neural layer followed by a BN layer
        if (type(layer) in supported_layers and
                len(layer.outbound_nodes) == 1 and
                isinstance(layer.outbound_nodes[0].layer, BatchNormalization)):

            layer_bn = layer.outbound_nodes[0].layer
            # To be folded, BN layer must have only one inbound node and
            # no more than one outbound node.
            bn_outbounds = layer_bn.outbound_nodes
            if (len(layer_bn.inbound_nodes) != 1 or len(bn_outbounds) > 1):
                continue

            map_neural_layer_to_bn[layer] = layer_bn
            if len(bn_outbounds) == 1:
                map_bn_to_next_layer[layer_bn] = bn_outbounds[0].layer

    return map_neural_layer_to_bn, map_bn_to_next_layer


def _prepare_model_to_fold_BN(model, map_neural_layer_to_bn,
                              map_bn_to_next_layer):
    """Prepare model to be cloned with folded BatchNormalization layers.

    To fold BN layers by using "clone_model", the model must be prepared by
    modifying some internal variables, such as _layers, _output_layers or
    _inbound_nodes.

    Three operations are done here:
    1. Remove BN layers from model.layers.
    2. Bypass BN layers in the graph by updating inbound nodes of the layers
       following the BN layers. The new inbound nodes are the neural layers
       preceding the BN layers.
    3. If a BN layer is an output layer of the model, the preceding neural
       layer must be added to the new output layers.

    The model instance and the layers are directly modified in this function.
    """

    # Remove BN layers from model.layers.
    for layer_bn in map_neural_layer_to_bn.values():
        model._layers.remove(layer_bn)  # pylint: disable=protected-access

    # Update inbound nodes as there is no BN between neural layer and
    # following layer. Inbound nodes are used in "clone_layer" function to
    # create the model architecture. Here, we replace the inbound node of the
    # next layer with the inbound node of the BN layer, in order to bypass the
    # BN layer.
    for layer_bn, next_layer in map_bn_to_next_layer.items():
        # pylint: disable=protected-access
        node = layer_bn.outbound_nodes[0]
        node_index = next_layer._inbound_nodes.index(node)
        next_layer._inbound_nodes[node_index] = layer_bn.inbound_nodes[0]

    # If BN layer is an output layer, replace it with its
    # inbound layer
    for neural_layer, layer_bn in map_neural_layer_to_bn.items():
        # pylint: disable=protected-access
        if layer_bn in model._output_layers:
            # Replace BN layer in _output_layers and _output_coordinates
            index_bn = model._output_layers.index(layer_bn)
            model._output_layers[index_bn] = neural_layer
            model._output_coordinates[index_bn] = (neural_layer, 0, 0)


def load_quantized_model(filepath, custom_objects=None, compile_model=True):
    """Loads a quantized model saved in TF or HDF5 format.

    If the model was compiled and trained before saving, its training state
    will be loaded as well.
    This function is a wrapper of `tf.keras.models.load_model`.

    Args:
        filepath (string): path to the saved model.
        custom_objects (dict): optional dictionary mapping names (strings) to
            custom classes or functions to be considered during deserialization.
        compile_model (bool): whether to compile the model after loading.

    Returns:
        :obj:`tensorflow.keras.Model`: a Keras model instance.
    """
    if custom_objects is None:
        custom_objects = {}
    all_objects = {**custom_objects, **cnn2snn_objects}
    return load_model(filepath, all_objects, compile_model)


def load_partial_weights(dest_model, src_model):
    """Loads a subset of weights from one Keras model to another

    This goes through each layers of the source model, looking for a matching
    layer in the destination model.
    If a layer with the same name is found, then this method assumes that one
    of the two layer has the same set of weights as the other plus some extra
    weights at the end, and loads only the first common weights into the
    destination layer.

    Args:
        dest_model(:obj:`tensorflow.keras.Model`): the destination Model
        src_model(:obj:`tensorflow.keras.Model`): the source Model

    """
    for src_layer in src_model.layers:
        src_weights = src_layer.get_weights()
        dest_layer = dest_model.get_layer(src_layer.name)
        dest_weights = dest_layer.get_weights()
        # Take the minimum of the two lists of weights
        n_weights = min(len(src_weights), len(dest_weights))
        # Only replace the first weights
        dest_weights[0:n_weights] = src_weights[0:n_weights]
        dest_layer.set_weights(dest_weights)


def create_trainable_quantizer_model(quantized_model):
    """Converts a legacy quantized model to a model using trainable quantizers.

    Legacy cnn2snn models have fixed quantization schemes. This method converts
    such a model to an equivalent model using trainable quantizers.

    Args:
        quantized_model(str, :obj:`tensorflow.keras.Model`): a keras Model or a
        path to a keras Model file

    Returns:
        :obj:`tensorflow.keras.Model`: a Keras Model instance.
    """
    if isinstance(quantized_model, str):
        # Load the model at the specified path
        quantized_model = load_quantized_model(quantized_model)
    # Dump model configuration in a JSON string
    json_string = quantized_model.to_json()
    # Edit the model configuration to replace static quantizers by trainable
    # ones
    json_string = json_string.replace("StdWeightQuantizer",
                                      "TrainableStdWeightQuantizer")
    json_string = json_string.replace("ActivationDiscreteRelu", "QuantizedReLU")
    # Create a new model from the modified configuration
    new_model = model_from_json(json_string, custom_objects=cnn2snn_objects)
    # Transfer weights to the new model
    load_partial_weights(new_model, quantized_model)
    return new_model


def chain_cloned_layer(x, layer):
    """This function creates a hard copy of a layer and apply to it the
    tf.Tensor x.

    """

    config = layer.get_config()
    w = layer.get_weights()
    new_layer = type(layer).from_config(config)
    out = new_layer(x)
    new_layer.set_weights(w)
    return out



# ******************************************************************************
# Copyright 2020 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""CNN2SNN toolkit to quantize and convert Keras models into Akida-compatible
models.
"""

from .utils import (load_quantized_model, load_partial_weights,
                    create_trainable_quantizer_model)

from .cnn2snn_objects import cnn2snn_objects
from .converter import convert
from .compatibility_checks import check_model_compatibility
from .quantization_ops import (StdWeightQuantizer, TrainableStdWeightQuantizer,
                               MaxQuantizer, MaxPerAxisQuantizer)
from .quantization_layers import (QuantizedConv2D, QuantizedDepthwiseConv2D,
                                  QuantizedDense, QuantizedSeparableConv2D,
                                  ActivationDiscreteRelu, QuantizedReLU,
                                  QuantizedActivation)
from .quantization import quantize, quantize_layer



# ******************************************************************************
# Copyright 2020 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Command-line interface"""

import argparse
import re
import os

from .converter import convert
from .quantization import quantize
from .utils import load_quantized_model
from .compatibility import _upgrade_model_with_legacy_quantizers


def quantize_model(model_path, wq, aq, iq):
    """Wrapper to quantize model"""
    model = load_quantized_model(model_path)
    if iq == -1:
        iq = wq
    model_q = quantize(model, wq, aq, iq)
    # Extract base name
    base_name = os.path.splitext(model_path)[0]
    # Cross-platform path may contain alpha characters, /, \ and :
    path_re = r"([\w/\\:]+)"
    # Quantization suffix has a well-identified structure
    suffix_re = r"(_iq\d_wq\d_aq\d)"
    p = re.compile(path_re + suffix_re)
    # Look for an existing quantization suffix in the base name
    m = p.match(base_name)
    if m:
        # Only keep the actual base name (group(2) contains the suffix)
        base_name = m.group(1)
    out_path = f"{base_name}_iq{iq}_wq{wq}_aq{aq}.h5"
    model_q.save(out_path, include_optimizer=False)
    print(f"Model successfully quantized and saved as {out_path}.")


def convert_model(model_path, scale, shift, is_image):
    """Wrapper to convert model"""
    base_name = os.path.splitext(model_path)[0]
    q_model = load_quantized_model(model_path)
    ak_model = convert(q_model,
                       input_scaling=(scale, shift),
                       input_is_image=is_image)
    out_path = f"{base_name}.fbz"
    ak_model.save(out_path)
    print(f"Model successfully converted and saved as {out_path}.")


def main():
    """CNN2SNN command-line interface to quantize/convert/upgrade a model"""
    parser = argparse.ArgumentParser()
    parser.add_argument("-m", "--model", default=None, help="The source model")
    sp = parser.add_subparsers(dest="action")
    q_parser = sp.add_parser("quantize", help="Quantize a Keras model")
    q_parser.add_argument("-wq",
                          "--weight_quantization",
                          type=int,
                          default=4,
                          help="The weight quantization")
    q_parser.add_argument("-aq",
                          "--activ_quantization",
                          type=int,
                          default=4,
                          help="The activations quantization")
    q_parser.add_argument("-iq",
                          "--input_weight_quantization",
                          type=int,
                          default=-1,
                          help="The first layer weight quantization (same as" \
                          " weight_quantization if omitted)")
    c_parser = sp.add_parser(
        "convert", help="Convert a quantized Keras model to an akida model")
    c_parser.add_argument("-sc",
                          "--scale",
                          type=int,
                          default=1,
                          help="The scale factor applied on uint8 inputs.")
    c_parser.add_argument("-sh",
                          "--shift",
                          type=int,
                          default=0,
                          help="The shift applied on uint8 inputs.")
    c_parser.add_argument("--no-image-input",
                          action='store_true',
                          default=False,
                          help="If inputs are not 8-bit images")
    u_parser = sp.add_parser(
        "upgrade", help="Upgrade model to be compatible with cnn2snn>=1.8.10")
    u_parser.add_argument("-o",
                          "--output",
                          type=str,
                          default=None,
                          help="Output model filepath")

    args = parser.parse_args()
    if args.action == "quantize":
        quantize_model(args.model, args.weight_quantization,
                       args.activ_quantization, args.input_weight_quantization)
    if args.action == "convert":
        convert_model(args.model, args.scale, args.shift,
                      not args.no_image_input)
    if args.action == "upgrade":
        _upgrade_model_with_legacy_quantizers(args.model, args.output)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2021 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to check model compatibility for CNN2SNN conversion.
"""

from tensorflow.keras import Sequential, layers

from .transforms.sequential import (_raise_error_if_model_not_sequential,
                                    cut_model_to_sequentials,
                                    transform_submodels)

from . import quantization_layers as qlayers

neural_layers = (layers.Conv2D, layers.SeparableConv2D, layers.Dense,
                 qlayers.QuantizedConv2D, qlayers.QuantizedSeparableConv2D,
                 qlayers.QuantizedDense)


def check_model_compatibility(model, input_is_image=True):
    r"""Checks if a Keras model is compatible for cnn2snn conversion.

    This function doesn't convert the Keras model to an Akida model
    but only checks if the model design is compatible.

    Note that this function doesn't check if the quantization bitwidths (weights
    or activations) are supported by the Akida Execution Engine or by the Akida
    NSoC.

    **1. How to build a compatible Keras quantized model?**

    The following lines give details and constraints on how to build a Keras
    model compatible for the conversion to an Akida model.


    **2. General information about layers**

    An Akida layer must be seen as a block of Keras layers starting with a
    processing layer (Conv2D, SeparableConv2D,
    Dense). All blocks of Keras layers except the last block must have
    exactly one activation layer (ReLU or ActivationDiscreteRelu). Other
    optional layers can be present in a block such as a pooling layer or a
    batch normalization layer.
    Here are all the supported Keras layers for an Akida-compatible model:

    - Processing layers:

      - tf.keras Conv2D/SeparableConv2D/Dense
      - cnn2snn QuantizedConv2D/QuantizedSeparableConv2D/QuantizedDense

    - Activation layers:

      - tf.keras ReLU
      - cnn2snn ActivationDiscreteRelu
      - any increasing activation function (only for the last block of layers)
        such as softmax, sigmoid set as last layer. This layer must derive from
        tf.keras.layers.Activation, and it will be removed during conversion to
        an Akida model.

    - Pooling layers:

      - MaxPool2D
      - GlobalAvgPool2D

    - BatchNormalization
    - Dropout
    - Flatten
    - Input
    - Reshape

    Example of a block of Keras layers::

              ----------
              | Conv2D |
              ----------
                  ||
                  \/
        ----------------------
        | BatchNormalization |
        ----------------------
                  ||
                  \/
             -------------
             | MaxPool2D |
             -------------
                  ||
                  \/
       --------------------------
       | ActivationDiscreteRelu |
       --------------------------


    **3. Constraints about inputs**

    An Akida model can accept two types of inputs: sparse events or 8-bit
    images. Whatever the input type, the Keras inputs must respect the
    following relation:

        input_akida = scale * input_keras + shift

    where the Akida inputs must be positive integers, the input scale must be
    a float value and the input shift must be an integer. In other words,
    scale * input_keras must be integers.

    Depending on the input type:

    - if the inputs are events (sparse), the first layer of the Keras model can
      be any processing layer. The input shift must be zero.
    - if the inputs are images, the first layer must be a Conv2D
      layer.


    **4. Constraints about layers' parameters**

    To be Akida-compatible, the Keras layers must observe the following rules:

    - all layers with the 'data_format' parameter must be 'channels_last'
    - all processing quantized layers and ActivationDiscreteRelu must have a
      valid quantization bitwidth
    - a Dense layer must have an input shape of (N,) or (1, 1, N)
    - a BatchNormalization layer must have 'axis' set to -1 (default)
    - a BatchNormalization layer cannot have negative gammas
    - Reshape layers can only be used to transform a tensor of shape (N,) to a
      tensor of shape (1, 1, N), and vice-versa
    - only one pooling layer can be used in each block
    - a MaxPool2D layer must have the same 'padding' as the corresponding
      processing quantized layer

    **5. Constraints about the order of layers**

    To be Akida-compatible, the order of Keras layers must observe the following
    rules:

    - a block of Keras layers must start with a processing quantized layer
    - where present, a BatchNormalization/GlobalAvgPool2D layer must be placed
      before the activation
    - a Flatten layer can only be used before a Dense layer
    - an Activation layer other than ActivationDiscreteRelu can only be used
      in the last layer


    Args:
        model (:obj:`tf.keras.Model`): the model to check.
        input_is_image (bool, optional): True if input is an image (8-bit input
            with 1 or 3 channels) followed by QuantizedConv2D. Akida model
            input will be InputConvolutional. If False, Akida model input will
            be InputData. (Default value = True)
    """
    try:
        model_cut = cut_model_to_sequentials(model)
        model_transform = transform_submodels(model_cut)
        check_functional_compatibility(model_transform, input_is_image)
        return True
    except RuntimeError as e:
        print(
            "The Keras quantized model is not compatible for a conversion "
            "to an Akida model:\n", str(e))
        return False


def check_functional_compatibility(model, input_is_image):
    """Checks compatibility of a functional Keras model.

    The expected model to check must have been previously cut into Sequential
    submodels and transformed using `cut_model_to_sequentials` and
    `transform_submodels`.

    The model is parsed and each sequential submodel is checked for
    compatibility. If an incompatibility is detected, an error is raised. This
    function can be applied on native Keras models and does not check
    quantization parameters.

    Args:
        model (:obj:`tf.keras.Model`): a Functional or Sequential Keras model.
        input_is_image (bool): True if input is an image.
    """

    if isinstance(model, Sequential):
        sequential_check_compatibility(model, input_is_image)
        return

    _check_outbounds_of_input_layer(model, input_is_image)

    def parse_submodel(layer, input_is_image, visited_layers):
        # pylint: disable=duplicate-code
        if layer in visited_layers:
            raise RuntimeError(f"Layer {layer.name} already visited.")

        # Do not visit this layer if all inbound layers were not visited yet
        inbound_layers = [n.layer for n in layer.inbound_nodes[0].parent_nodes]
        if set(inbound_layers).difference(visited_layers):
            return
        visited_layers.append(layer)

        if isinstance(layer, layers.InputLayer):
            pass
        elif isinstance(layer, layers.Concatenate):
            pass
        elif isinstance(layer, Sequential):
            # Get previous Akida layer of this submodel
            is_last_submodel = (layer.name == last_submodel)
            sequential_check_compatibility(layer, input_is_image,
                                           is_last_submodel)
            input_is_image = None
        else:
            raise RuntimeError(f"Layer {layer.name} of type "
                               f"{layer.__class__.__name__} is not supported.")

        # Go to next layer/submodel
        for next_layer in [node.layer for node in layer.outbound_nodes]:
            parse_submodel(next_layer, input_is_image, visited_layers)

    # Go through model layers to generate Akida layers
    input_layer = model.get_layer(model.input_names[0])
    last_submodel = model.layers[-1].name
    parse_submodel(input_layer, input_is_image, visited_layers=[])


def sequential_check_compatibility(model,
                                   input_is_image,
                                   is_last_submodel=True):
    """Checks Sequential (sub)model compatibility.

    If an incompatibility is detected, an error is raised. This function can be
    applied on native Keras models and does not check quantization parameters.

    Args:
        model (:obj:`tf.keras.Model`): a Sequential Keras model.
        input_is_image (bool): True if input is an image, False if input is
            events, None if the submodel is not the first submodel of the
            functional model.
        is_last_submodel (bool): True if the submodel is the last one of the
            functional model.
    """

    _raise_error_if_model_not_sequential(model)
    if len(model.layers) == 0:
        return

    # Check if the first layer is a (Quantized)Conv2D if input_is_image=True
    if input_is_image:
        _check_first_layer_with_image_input(model)

    # The two following variables are used to know if there is an alternation
    # between a neural layer and an activation layer. For example, when we go
    # through a neural layer in the model, 'last_visited_neural_layer' must be
    # None. If not, it means that there is no activation layer between this
    # neural layer and the previous one. 'last_visited_neural_layer' is then set
    # to the current neural layer and 'last_visited_activation' is set to None.
    # A (sub)model must start with a neural layer.
    last_visited_neural_layer = None
    last_visited_activation = model.layers[0]

    for i, layer in enumerate(model.layers):
        if type(layer) in neural_layers:
            _check_dense_shape(layer)
            _check_two_neural_layers(layer, last_visited_neural_layer)
            # Update last visited neural and activation layer
            last_visited_neural_layer = layer
            last_visited_activation = None
        elif isinstance(layer, (layers.ReLU, qlayers.QuantizedActivation)):
            _check_two_successive_activations(layer, last_visited_activation)
            # Update last visited neural and activation layer
            last_visited_neural_layer = None
            last_visited_activation = layer
        elif isinstance(layer, (layers.MaxPool2D, layers.GlobalAvgPool2D)):
            _check_pooling_compatibility(model, i)
        elif isinstance(layer, layers.Flatten):
            _check_flatten_layer(model, i)
        elif isinstance(layer, layers.Reshape):
            _check_reshape_layer(layer)
        # Rescaling layer only supported as first layer of the first submodel
        elif isinstance(layer, layers.experimental.preprocessing.Rescaling):
            _check_rescaling_compatibility(layer, i, input_is_image)
        # Activation/Softmax layer only supported after the last neural layer of
        # the last submodel
        elif isinstance(layer, (layers.Activation, layers.Softmax)):
            _check_last_activation(model, i, is_last_submodel)
        else:
            raise RuntimeError(
                f"Layer {layer.name} of type {layer.__class__.__name__} is not "
                f"supported for Akida conversion")

    # Check if any submodel other than the last one ends with an activation
    # rather than with a neural layer
    if not is_last_submodel and last_visited_activation is None:
        raise RuntimeError(f"An activation layer must be present after the "
                           f"neural layer '{last_visited_neural_layer.name}'.")


def _check_outbounds_of_input_layer(model, input_is_image):
    """Asserts that an input layer with input_is_image=True must have only one
    outbound layer. The corresponding Akida model will start with an
    InputConvolutional layer.
    """
    if not input_is_image:
        return

    outbounds = model.get_layer(model.input_names[0]).outbound_nodes
    if len(outbounds) > 1:
        outbound_layers = [node.layer.layers[0].name for node in outbounds]
        raise RuntimeError("With input_is_image=True, the model must not start "
                           "with a split into two layers. Receives layers "
                           f"{outbound_layers}.")


def _check_first_layer_with_image_input(model):
    """Checks that the first neural layer of a model with input_is_image=True
    is a Conv2D or QuantizedConv2D layer.

    Args:
        model (:obj:`tf.keras.Model`): a Sequential Keras (sub)model.
    """

    next_layer = 0
    if isinstance(model.layers[next_layer],
                  layers.experimental.preprocessing.Rescaling):
        next_layer += 1

    if (type(model.layers[next_layer])
            not in (layers.Conv2D, qlayers.QuantizedConv2D) or
            model.layers[next_layer].input_shape[-1] not in (1, 3)):
        layer = model.layers[next_layer]
        raise RuntimeError(
            f"With input_is_image=True, first layer '{layer.name}' must be "
            f"Conv2D and input shape must have 1 or 3 channels. Receives layer "
            f"of type {layer.__class__.__name__} with {layer.input_shape[-1]} "
            f"channels.")


def _check_dense_shape(layer):
    """Asserts Dense layer is compatible for conversion.
    One check is performed here:
    - input shape must be (bs, N) or (bs, 1, 1, N) (bs is the batch size).

    Args:
        layer(:obj:`tf.keras.Layer`): the Dense or QuantizedDense layer to
            check.
    """

    if type(layer) not in (layers.Dense, qlayers.QuantizedDense):
        return

    # Raises error if Dense input shape is incorrect: supported
    # shapes are (N,) and (1, 1, N). Remember input_shape has the batch
    # size as first element of tuple.
    valid = (  # Input shape is (N,)
        len(layer.input_shape) == 2 or
        # Input shape is (1, 1, N)
        (len(layer.input_shape) == 4 and layer.input_shape[1] == 1 and
         layer.input_shape[2] == 1))
    if not valid:
        raise RuntimeError(f"The Dense layer {layer.name} must have an "
                           "input shape of (N,) or (1,1,N). Receives "
                           f"{layer.input_shape[1:]}.")


def _check_two_neural_layers(layer, last_visited_neural_layer):
    """Checks that there is an activation between two neural layers.

    Args:
        layer (:obj:`tf.keras.Layer`): the current neural layer.
        last_visited_neural_layer (:obj:`tf.keras.Layer`): the last visited
            neural layer if there is no activation between this last visited and
            the current neural layer. If there is an activation, this argument
            must be None.
    """

    if last_visited_neural_layer is not None:
        raise RuntimeError(
            f"An activation layer is required between the two neural layers "
            f"'{last_visited_neural_layer.name}' and '{layer.name}'.")


def _check_two_successive_activations(layer, last_visited_activation):
    """Checks that there is a neural layer between two activation layers.

    Args:
        layer (:obj:`tf.keras.Layer`): the current activation layer.
        last_visited_activation (:obj:`tf.keras.Layer`): the last visited
            activation layer if there is no neural layer between this last
            visited and the current activation. If there is a neural layer, this
            argument must be None.
    """
    if last_visited_activation is not None:
        raise RuntimeError(
            f"A neural layer is required between the two activation layers "
            f"'{last_visited_activation.name}' and '{layer.name}'.")


def _check_pooling_compatibility(model, pool_index):
    """Asserts pooling layer is compatible for conversion. Transformations must
    have beed applied before these checks.
    Two checks are performed here:
    - a pooling layer must be placed directly after a neural layer
    - the padding of MaxPool2D must be the same as the padding of neural layer

    Args:
        model (:obj:`tf.keras.Model`): the Sequential model to check.
        pool_index (int): the index of the pooling layer to check.
    """

    conv_neural_layers = (layers.Conv2D, layers.SeparableConv2D,
                          qlayers.QuantizedConv2D,
                          qlayers.QuantizedSeparableConv2D)

    layer_pool = model.layers[pool_index]
    prev_layer = model.layers[pool_index - 1]

    # Raise error if GlobalAvgPool2D is placed after the activation
    if type(prev_layer) not in conv_neural_layers:
        raise RuntimeError(f"Pooling layer {layer_pool.name} must be placed "
                           "after a convolutional neural layer. Currently after"
                           f" {prev_layer.name}.")

    # Raises error if the padding of MaxPool2D is different from the padding
    # of the neural processing layer.
    if (isinstance(layer_pool, layers.MaxPool2D) and
            prev_layer.padding != layer_pool.padding):
        raise RuntimeError(f"Pooling layer {layer_pool.name} (padding: "
                           f"{layer_pool.padding}) must have the same "
                           f"padding as {prev_layer.name} (padding: "
                           f"{prev_layer.padding}).")


def _check_flatten_layer(model, flatten_index):
    """Checks that the Flatten layer is supported for conversion.
    A Flatten layer is supported if it is followed by a Dense layer or if it
    is the last layer of the model.

    Args:
        model (:obj:`tf.keras.Model`): a Sequential Keras (sub)model.
        flatten_index (int): the index of the Flatten layer in the model's
            layers.
    """

    try:
        if type(model.layers[flatten_index +
                             1]) not in (layers.Dense, qlayers.QuantizedDense):
            raise RuntimeError(
                "A Flatten layer is only supported before a Dense one. Receives"
                f" unsupported Flatten layer '{model.layers[flatten_index]}'.")
    except IndexError:
        pass


def _check_reshape_layer(layer):
    """This function checks if the Reshape layer is supported.
    In the cnn2snn conversion, a Reshape layer can only be used to transform
    a tensor of shape (N,) to a tensor of shape (1, 1, N), and vice-versa.
    Note that the 'input_shape' and 'output_shape' parameters of a layer has
    the batch size as first element:
        input_shape = (batch_size,) + input_tensor_shape
    The batch size is ignored in the following function.

    Args:
        layer(:obj:`tf.keras.Layer`): the Reshape layer to check.
    """
    in_shape = layer.input_shape
    out_shape = layer.output_shape

    valid = ((  # Reshape from (1,1,N) to (N,)
        len(in_shape) == 4 and in_shape[1] == 1 and in_shape[2] == 1 and
        len(out_shape) == 2 and out_shape[1] == in_shape[3]) or
             # Reshape from (N,) to (1,1,N)
             (len(in_shape) == 2 and len(out_shape) == 4 and
              out_shape[1] == 1 and out_shape[2] == 1 and
              out_shape[3] == in_shape[1]) or
             # Useless Reshape, from X to X
             (in_shape == out_shape))

    if not valid:
        raise RuntimeError(f"The Reshape layer {layer.name} can only be used "
                           "to transform a tensor of shape (N,) to a tensor of "
                           "shape (1, 1, N), and vice-versa. Receives "
                           f"input_shape {in_shape[1:]} and output_shape "
                           f"{out_shape[1:]}.")


def _check_rescaling_compatibility(layer, index, input_is_image):
    """Checks that Rescaling layer is only accepted as the first layer of the
    first submodel.

    Args:
        layer(:obj:`tf.keras.Layer`): the Rescaling layer to check.
        index (int): the index of the Rescaling layer in the model.
        input_is_image (bool): True or False if the submodel is the first one of
            the functional model. None if the submodel is not the first one.
    """

    if input_is_image is None or index != 0:
        raise RuntimeError("A Rescaling layer must be the first layer of the "
                           f"model. Receives unexpected layer {layer.name}.")


def _check_last_activation(model, index, is_last_submodel):
    """Checks that an unsupported activation layer (tf.keras.Activation or
    tf.keras.Softmax) is only present after the last neural layer of the last
    submodel.

    Args:
        model(:obj:`tf.keras.Model`): the (sub)model to check.
        index (int): the index of the unsupported activation layer in the model.
        is_last_submodel (bool): True if the submodel is the last one of the
            the functional model.
    """

    last_layers_type = {type(layer) for layer in model.layers[index + 1:]}
    if not is_last_submodel or last_layers_type.intersection(neural_layers):
        raise RuntimeError(
            "Activation layers other than ReLU and quantized activations are "
            "not supported before the last neural layer. Receives activation "
            f"layer '{model.layers[index].name}' before a neural layer.")



from os import path
from setuptools import setup

# Read the contents of the README file
directory = path.abspath(path.dirname(__file__))
with open(path.join(directory, 'README'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='cnn2snn',
    version='1.9.0',
    description='Keras to Akida CNN Converter',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Alvaro Moran',
    author_email='amoran@brainchip.com',
    url='https://doc.brainchipinc.com',
    license='Apache 2.0',
    packages=['cnn2snn', 'cnn2snn.transforms'],
    entry_points={
        'console_scripts': [ 'cnn2snn = cnn2snn.cli:main' ]
    },
    install_requires=['akida==1.9.0', 'tensorflow~=2.4.0', 'numpy'],
)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2021 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Transformation utilities for Keras/CNN2SNN Sequential models.

These transformations are designed for a specific stage in the CNN2SNN
conversion workflow. They must not be used as a generic tool to transform any
Sequential models.
"""

import numpy as np
import tensorflow as tf
from tensorflow.keras import Model, Sequential, Input
from tensorflow.keras.layers import (Conv2D, SeparableConv2D, Dense, MaxPool2D,
                                     GlobalAvgPool2D, ReLU, BatchNormalization,
                                     Dropout, Concatenate, InputLayer)

from ..cnn2snn_objects import cnn2snn_objects
from ..quantization_ops import ScaleFactorQuantizer
from ..quantization_layers import (QuantizedConv2D, QuantizedSeparableConv2D,
                                   QuantizedDense, QuantizedActivation)


def _clone_layer(layer):
    return layer.__class__.from_config(layer.get_config())


def _clone_layer_and_add_to_model(layer, model):
    layer_clone = _clone_layer(layer)
    model.add(layer_clone)
    layer_clone.set_weights(layer.get_weights())
    return layer_clone


def sequential_remove_useless_layers(model):
    """Removes useless layers in a Sequential model.

    Useless layers are:
    - Dropout

    Args:
        model (:obj:`tf.keras.Model`): a Sequential Keras model.
    Returns:
        :obj:`tf.keras.Model`: a Sequential Keras model.
    """

    _raise_error_if_model_not_sequential(model)

    useless_layers = (Dropout,)

    new_model = Sequential()
    new_model.add(Input(model.input_shape[1:]))
    for layer in model.layers:
        if isinstance(layer, useless_layers):
            continue
        _clone_layer_and_add_to_model(layer, new_model)

    return new_model


def sequential_invert_pooling_activation(model):
    """Invert activation and MaxPool2D layer in a Sequential model to have
    MaxPool2D before the activation.

    Having activation->MaxPool2D or MaxPool2D->activation is equivalent only if
    the activation is increasing (ok for ReLU and QuantizedActivation).
    Note that GlobalAvgPool2D cannot be inverted with an activation because
    there is no equivalence between activation->GAP and GAP->activation.

    Args:
        model (:obj:`tf.keras.Model`): a Sequential Keras model.
    Returns:
        :obj:`tf.keras.Model`: a Sequential Keras model.
    """

    _raise_error_if_model_not_sequential(model)

    new_model = Sequential()
    new_model.add(Input(model.input_shape[1:]))
    i = 0
    while i < len(model.layers) - 1:
        layer = model.layers[i]
        next_layer = model.layers[i + 1]
        if (isinstance(layer, (ReLU, QuantizedActivation)) and
                isinstance(next_layer, MaxPool2D)):
            _clone_layer_and_add_to_model(next_layer, new_model)
            _clone_layer_and_add_to_model(layer, new_model)
            i = i + 2
        else:
            _clone_layer_and_add_to_model(layer, new_model)
            i = i + 1

    if i < len(model.layers):
        _clone_layer_and_add_to_model(model.layers[-1], new_model)

    return new_model


def sequential_invert_batchnorm_pooling(model):
    """Inverts pooling and BatchNormalization layers in a Sequential model to
    have BN layer before pooling.

    Having pool->BN or BN->pool is equivalent only if BN layer has no negative
    gammas.

    Args:
        model (:obj:`tf.keras.Model`): a Sequential Keras model.
    Returns:
        :obj:`tf.keras.Model`: a Sequential Keras model.
    """

    _raise_error_if_model_not_sequential(model)

    new_model = Sequential()
    new_model.add(Input(model.input_shape[1:]))
    i = 0
    while i < len(model.layers) - 1:
        layer = model.layers[i]
        next_layer = model.layers[i + 1]
        if (isinstance(layer, (MaxPool2D, GlobalAvgPool2D)) and
                isinstance(next_layer, BatchNormalization)):
            gammas = next_layer.get_weights()[0]
            if isinstance(layer, MaxPool2D) and np.any(gammas <= 0):
                # It is impossible to invert MaxPool->BN with gammas <= 0
                raise RuntimeError(f"There are {np.sum(gammas <= 0)} negative "
                                   "gammas in the batch norm layer "
                                   f"{next_layer.name}. Negative gammas are "
                                   "not supported.")
            # GlobalAveragePooling2D brings a change on axis for the batch norm.
            if isinstance(layer, GlobalAvgPool2D):
                bn_config = next_layer.get_config()
                bn_config['axis'] = [-1]
                bn_layer_clone = BatchNormalization.from_config(bn_config)
            else:
                bn_layer_clone = _clone_layer(next_layer)
            new_model.add(bn_layer_clone)
            bn_layer_clone.set_weights(next_layer.get_weights())
            _clone_layer_and_add_to_model(layer, new_model)
            i = i + 2
        else:
            _clone_layer_and_add_to_model(layer, new_model)
            i = i + 1

    if i < len(model.layers):
        _clone_layer_and_add_to_model(model.layers[-1], new_model)

    return new_model


def sequential_freeze_quantizers(model):
    """Freezes weight quantizers in quantized neural layers.

    Freezing weight quantizers consists in replacing any quantizer with a
    custom quantizer that directly stores the scale factors, instead of
    computing them depending on the weights. After freezing quantizers,
    the neural weights must not be modified anymore.
    Doing this allows to prepare any model for BatchNormalization folding,
    which is not compatible with all quantizers.

    Note that this operation is irreversible and must only be used in the
    conversion workflow.
    """

    _raise_error_if_model_not_sequential(model)

    neural_quantized_layers = (QuantizedConv2D, QuantizedSeparableConv2D,
                               QuantizedDense)

    new_model = Sequential()
    new_model.add(Input(model.input_shape[1:]))
    for layer in model.layers:
        layer_clone = _clone_layer_and_add_to_model(layer, new_model)
        if type(layer) in neural_quantized_layers:
            # Get bitwidth and weights scale factors
            bitwidth = layer.quantizer.bitwidth
            shift_for_sepconv = isinstance(layer, QuantizedSeparableConv2D)
            w = layer.get_weights()[0 + shift_for_sepconv]
            scale_factors = layer.quantizer.scale_factor(tf.constant(w))

            # Freeze quantizer by replacing with ScaleFactorQuantizer
            layer_clone.quantizer = ScaleFactorQuantizer(
                bitwidth, scale_factors)

    return new_model


def sequential_fold_BN(model):
    """Folds BatchNormalization layers into the preceding neural layers of
    a Sequential model.

    Args:
        model (:obj:`tf.keras.Model`): a Sequential Keras model.
    Returns:
        :obj:`tf.keras.Model`: a Sequential Keras model.
    """

    _raise_error_if_model_not_sequential(model)

    neural_layers = (Conv2D, SeparableConv2D, Dense, QuantizedConv2D,
                     QuantizedSeparableConv2D, QuantizedDense)

    new_model = Sequential()
    new_model.add(Input(model.input_shape[1:]))
    i = 0
    while i < len(model.layers) - 1:
        layer = model.layers[i]
        next_layer = model.layers[i + 1]
        if (isinstance(layer, neural_layers) and
                isinstance(next_layer, BatchNormalization)):

            # Check BN axis parameter
            if (len(next_layer.axis) != 1 or
                    next_layer.axis[0] != len(next_layer.input_shape) - 1):
                raise RuntimeError(f"The BatchNormalization layer "
                                   f"{next_layer.name} must be applied on the "
                                   f"last axis. Receives {next_layer.axis}.")

            # Add new neural layer with bias
            config = layer.get_config()
            config['use_bias'] = True
            new_layer = layer.__class__.from_config(config)
            new_model.add(new_layer)
            new_weights, scale_BN = compute_BN_folded_weights(layer, next_layer)
            if np.any(scale_BN == 0):
                # Zero gammas are not supported: once folded, new kernel is zero
                raise RuntimeError(f"There are {np.sum(scale_BN == 0)} null "
                                   "gammas in the batch norm layer "
                                   f"{next_layer.name}. Null gammas are not "
                                   "supported.")
            new_layer.set_weights(new_weights)

            # If the quantizer is ScaleFactorQuantizer, we must also update
            # the scale factors with the factor introduced by the BN
            if (hasattr(new_layer, 'quantizer') and
                    isinstance(new_layer.quantizer, ScaleFactorQuantizer)):
                new_layer.quantizer.scale_factors /= scale_BN
            i = i + 2
        else:
            _clone_layer_and_add_to_model(layer, new_model)
            i = i + 1

    if i < len(model.layers):
        _clone_layer_and_add_to_model(model.layers[-1], new_model)

    return new_model


def compute_BN_folded_weights(neural_layer, bn_layer):
    """Computes the new weights of a neural layer after folding BN layer.

    Args:
        neural_layer (:obj:`tf.keras.Layer`): a neural layer where BN will be
            folded.
        bn_layer (:obj:`tf.keras.Layer`): the BatchNormalization layer to fold
            into the neural layer.
    Returns:
        list: a list of the new weights to set in the new folded neural layer.
        list: a list of positive scale factors introduced by the folding.

    """

    # Get kernel and bias weights of the neural layer
    if type(neural_layer) in (SeparableConv2D, QuantizedSeparableConv2D):
        kernel_position = 1
        bias_position = 2
    else:
        kernel_position = 0
        bias_position = 1
    weights = neural_layer.get_weights()
    kernel = weights[kernel_position]
    bias = weights[bias_position] if neural_layer.use_bias else 0

    # Get BN weights
    gamma, beta, mean, var = bn_layer.get_weights()
    scale_BN = gamma / np.sqrt(var + bn_layer.epsilon)

    # Compute new folded kernel and bias
    new_kernel = kernel * scale_BN
    new_bias = beta + (bias - mean) * scale_BN

    # Return all weights with modified ones
    new_weights = weights
    new_weights[kernel_position] = new_kernel
    if neural_layer.use_bias:
        new_weights[bias_position] = new_bias
    else:
        new_weights.insert(bias_position, new_bias)

    # Absolute value of scale_BN is returned because we no longer need its sign.
    # It is later used to rescale the scale factors which are always positive.
    return new_weights, np.abs(scale_BN)


def _clone_model_with_weights(model):
    """Clones the model and copy weights into the cloned model.

    The cloned model is identical to the input model, but cloning removes the
    shared layers that can cause problems at inference.

    Args:
        model (:obj:`tf.keras.Model`): a Sequential Keras model.
    Returns:
        :obj:`tf.keras.Model`: a Sequential Keras model.
    """

    with tf.keras.utils.custom_object_scope(cnn2snn_objects):
        model_clone = tf.keras.models.clone_model(model)
    model_clone.set_weights(model.get_weights())
    return model_clone


def _raise_error_if_model_not_sequential(model):
    """Raises a ValueError if model is not a Sequential Keras model.
    """
    if not isinstance(model, Sequential):
        raise ValueError(f"Model is expected to be Sequential. Receives type "
                         f"{model.__class__.__name__}.")


def _check_layers_data_format(model):
    """Asserts that all layers in the model are 'channels_last'.
    Args:
        model (tf.keras.model): the Keras model to check.
    """

    # Error if 'data_format' is 'channels_first'
    for layer in model.layers:
        if hasattr(layer, 'data_format'):
            if layer.data_format == "channels_first":
                raise RuntimeError("Unsupported data format 'channels_first' "
                                   f"in layer '{layer.name}'.")


def _check_model_input_output(model):
    """Asserts that model inputs/outputs are supported for conversion.
    The Keras model must have only one input layer and one output layer. The
    input shape must 4-D (N, H, W, C).

    Args:
        model (tf.keras.model): the Keras model to check.
    """

    # Error if multiple inputs
    if len(model.input_names) > 1:
        raise RuntimeError("Model must have only one input layer. Receives "
                           f"inputs {model.input_names}.")

    # Error if multiple outputs
    if len(model.output_names) != 1:
        raise RuntimeError("Model must have only one output layer. Receives"
                           f"outputs {model.output_names}.")

    # Error if input shape is not 4D, i.e. (N, H, W, C)
    if len(model.input_shape) not in (2, 4):
        raise RuntimeError(
            "Input shape of model must be 2-D or 4-D (batch size + 1-D or 3-D "
            f"tensors). Receives input shape {model.input_shape}.")


def _check_layer_inbounds(model):
    """Asserts that all layers in the model have only one inbound node and
    inbound layer (except for input layers and Concatenate layers).
    """
    for layer in model.layers:
        if len(layer.inbound_nodes) != 1:
            raise RuntimeError("Layers must have only one inbound node. "
                               f"Receives layer {layer.name} with "
                               f"{len(layer.inbound_nodes)} nodes.")

        if (layer.inbound_nodes[0].is_input or isinstance(layer, Concatenate)):
            continue

        num_inbound_layers = len(layer.inbound_nodes[0].parent_nodes)
        if num_inbound_layers != 1:
            raise RuntimeError(f"Layer {layer.name} must have only one inbound "
                               f"layer. Receives {num_inbound_layers}.")


def _check_concat_layer_compatibility(layer, inbound_layers):
    """Checks that the Concatenate layer is compatible for conversion:

    - A Concatenate layer axis parameter must be equal to -1.
    - A Concatenate layer must have exactly two inbound layers.

    Args:
        layer (:obj:`tf.keras.Layer`): a Keras Concatenate layer.
        inbound_layers (list): the inbound layers of the Concatenate layer.
    """

    if layer.axis != -1:
        raise RuntimeError(f"The Concatenate layer '{layer.name}' must have "
                           f"axis=-1. Receives axis={layer.axis}.")

    if len(inbound_layers) != 2:
        inbound_names = [l.name for l in inbound_layers]
        raise RuntimeError(f"The Concatenate layer '{layer.name}' must have "
                           f"exactly two inbound layers. Receives inbound "
                           f"layers {inbound_names}.")


def cut_model_to_sequentials(model):
    """Cut a Keras model into Sequential sub-models and Concatenate layers.

    This function returns an equivalent model where all linear branches are
    replaced by a Sequential sub-model.

    Args:
        model (:obj:`tf.keras.Model`): a Keras model

    Returns:
        :obj:`tf.keras.Model`: a Keras model with Sequential sub-models
    """

    # Clone model to avoid shared layers
    model_clone = _clone_model_with_weights(model)

    _check_layers_data_format(model_clone)
    _check_model_input_output(model_clone)
    _check_layer_inbounds(model_clone)

    if isinstance(model_clone, Sequential):
        return model_clone

    def parse_layer(layer, visited_layers, current_branch, output_tensors):
        """Go through a TensorFlow/Keras graph by recursively looking at
        outbound layers.

        Each linear branch is detected and converted to a Sequential sub-model.
        A linear branch ends if:
          - the current layer has multiple outbounds (split connections)
          - the current layer has no outbound (output layer of the model)
          - the next layer is a Concatenate layer
        """

        if layer in visited_layers:
            raise RuntimeError(f"Layer {layer.name} already visited.")

        # Do not visit this layer if all inbound layers were not visited yet
        # (e.g. for Concatenate inbounds)
        # pylint: disable=duplicate-code
        inbound_layers = [n.layer for n in layer.inbound_nodes[0].parent_nodes]
        if set(inbound_layers).difference(visited_layers):
            return
        visited_layers.append(layer)

        # Skip input layer but store its output tensor for graph connection
        if isinstance(layer, InputLayer):
            output_tensors[layer] = layer.output
        # Add layer to graph if layer is Concatenate
        elif isinstance(layer, Concatenate):
            _check_concat_layer_compatibility(layer, inbound_layers)
            # Get input tensors of Concatenate layer
            concat_input_tensors = [output_tensors[l] for l in inbound_layers]
            # Add Concatenate layer to the graph
            output_tensors[layer] = layer(concat_input_tensors)

        else:
            # Add current layer to current branch
            current_branch.append(layer)

            # End current branch and add it to the graph if:
            # - current layer has multiple outbounds (split connections)
            # - current layer has no outbound (output layer of the model)
            # - next layer is Concatenate
            if (len(layer.outbound_nodes) != 1 or
                    isinstance(layer.outbound_nodes[0].layer, Concatenate)):

                parent_nodes = current_branch[0].inbound_nodes[0].parent_nodes
                input_tensor = output_tensors[parent_nodes[0].layer]

                # Create sub-model for current branch and add it to the graph
                submodel = Sequential(current_branch)
                output_tensors[layer] = submodel(input_tensor)
                current_branch.clear()

        # Go to next layer
        for next_layer in [node.layer for node in layer.outbound_nodes]:
            parse_layer(next_layer, visited_layers, current_branch,
                        output_tensors)

    # Go through model layers to detect Sequential branches
    input_layer = model_clone.get_layer(model_clone.input_names[0])
    output_layer = model_clone.get_layer(model_clone.output_names[0])
    output_tensors = {}
    parse_layer(input_layer,
                visited_layers=[],
                current_branch=[],
                output_tensors=output_tensors)

    # Create new model with Sequential sub-models
    model_cut = Model(model_clone.input, output_tensors[output_layer])

    # Clone model to avoid shared layers
    model_cut = _clone_model_with_weights(model_cut)

    return model_cut


def transform_submodels(model):
    """Transforms all Sequential submodels of a functional model.

    The input model must be composed of Sequential submodels and Concatenate
    layers. This function will apply transformations on every Sequential
    submodel and returns an equivalent functional model with transformed
    submodels.

    Args:
        model (:obj:`tf.keras.Model`): a Keras model with Sequential submodels.

    Returns:
        :obj:`tf.keras.Model`: a Keras model with transformed Sequential
            submodels.
    """

    def transform_sequential(submodel):
        submodel = sequential_remove_useless_layers(submodel)
        submodel = sequential_invert_pooling_activation(submodel)
        submodel = sequential_invert_batchnorm_pooling(submodel)
        submodel = sequential_freeze_quantizers(submodel)
        submodel = sequential_fold_BN(submodel)
        return submodel

    if isinstance(model, Sequential):
        return transform_sequential(model)

    def parse_submodel(layer, visited_layers, output_tensors):
        if layer in visited_layers:
            raise RuntimeError(f"Layer {layer.name} already visited.")

        # Do not visit this layer if all inbound layers were not visited yet
        # (e.g. for Concatenate inbounds)
        # pylint: disable=duplicate-code
        inbound_layers = [n.layer for n in layer.inbound_nodes[0].parent_nodes]
        if set(inbound_layers).difference(visited_layers):
            return
        visited_layers.append(layer)

        # Add layer to graph if layer is InputLayer or Concatenate
        if isinstance(layer, InputLayer):
            input_clone = _clone_layer(layer)
            output_tensors[layer] = input_clone.output
        elif isinstance(layer, Concatenate):
            # Get input tensors of Concatenate layer
            concat_input_tensors = [output_tensors[l] for l in inbound_layers]
            # Add Concatenate layer to the graph
            concat_clone = _clone_layer(layer)
            output_tensors[layer] = concat_clone(concat_input_tensors)
        elif isinstance(layer, Sequential):
            # Get input tensors of submodel
            parent_nodes = layer.inbound_nodes[0].parent_nodes
            input_tensor = output_tensors[parent_nodes[0].layer]
            # Transform submodel and add it to the graph
            new_submodel = transform_sequential(layer)
            output_tensors[layer] = new_submodel(input_tensor)
        else:
            # pylint: disable=duplicate-code
            raise RuntimeError(f"Layer {layer.name} of type "
                               f"{layer.__class__.__name__} is not supported "
                               f"here.")

        # Go to next layer
        for next_layer in [node.layer for node in layer.outbound_nodes]:
            parse_submodel(next_layer, visited_layers, output_tensors)

    # Go through model layers to transform Sequential branches
    input_layer = model.get_layer(model.input_names[0])
    output_tensors = {}
    parse_submodel(input_layer,
                   visited_layers=[],
                   output_tensors=output_tensors)

    # Create new model with transformed Sequential sub-models
    output_layer = model.get_layer(model.output_names[0])
    return Model(output_tensors[input_layer], output_tensors[output_layer])
